/*
  *	Linux Netfilter compability with IP Filter,To support iptables,ipchain etc.
  *													--Li Wang
  */

/*
  *	TO DO:
  *		1.When use mtx_lock() to simulate down_interruptible(),how to
  *		   return EINTR.turnstile_wait() return void.so mtx_lock() return void.
  *		2.Use sx_slock(),sx_xlock() to simulate read_lock()/write_lcok() 
  *		   is safe enough?the performance is whether better than mtx_lock()?
  *		3.Implement SMP_ALIGN()
  *		4.Need adding a field 'nfcache' to struct mbuf?
  *		5.Support runing as a loadable kernel module
  */
  
/* In this code, we can be waiting indefinitely for userspace to
 * service a packet if a hook returns NF_QUEUE.  We could keep a count
 * of skbuffs queued for userspace, and not deregister a hook unless
 * this is zero, but that sucks.  Now, we simply check when the
 * packets come back: if the hook is gone, the packet is discarded. */

#include <sys/cdefs.h>
__FBSDID("src/sys/compat/linux/linux_netfilter.c,v 1.160 2004/08/16 11:12:57 Li Wang Exp $");


#include <compat/linux/linux_netfilter.h>      /* Some structure definitions */

#ifdef CONFIG_NETFILTER_DEBUG
#define NFDEBUG(format, args...)  log(LOG_DEBUG, "Netfilter/Iptables:"format , ## args)
#else
#define NFDEBUG(format, args...)
#endif

#ifdef CONFIG_NETFILTER_DEBUG
#define IP_NF_ASSERT(x)						\
do {								\
	if (!(x))						\
		log(LOG_DEBUG, "IP_NF_ASSERT: %s:%s:%u\n",		\
		       __FUNCTION__, __FILE__, __LINE__);	\
} while(0)
#else
#define IP_NF_ASSERT(x)
#endif

#ifdef SMP
#define smp_num_cpus			smp_cpus
#define SMP_ALIGN(x)				x
#define smp_processor_id()		curthread->td_oncpu
#define TABLE_OFFSET(t,p) 		(SMP_ALIGN((t)->size)*(p))
#else
#define smp_num_cpus			1
#define SMP_ALIGN(x)				x
#define smp_processor_id()		0
#define TABLE_OFFSET(t,p) 		0
#endif

static inline int cpu_number_map(int cpu)
{
	return cpu;
}

#define FILTER_VALID_HOOKS ((1 << NF_IP_LOCAL_IN) | (1 << NF_IP_FORWARD) | (1 << NF_IP_LOCAL_OUT))

#ifndef sx_initialized
#define	sx_initialized(m)					((m)->sx_object.lo_flags & LO_INITIALIZED)
#endif

static struct mtx nf_sockopt_mutex;   /* Lock for Netfilter operations */

static struct list_head nf_sockopts;

static struct mtx ipt_mutex;

static struct list_head ipt_tables;

static struct list_head ipt_target;

static struct list_head ipt_match;

#define ADD_COUNTER(c,b,p) do { (c).bcnt += (b); (c).pcnt += (p); } while(0)

MALLOC_DEFINE(M_IPTBUF, "iptbuf", "Buffers for the Linux Netfilter Iptables");

static inline
int do_match(struct ipt_entry_match *m,
	     const struct mbuf *skb,
	     const struct ifnet *in,
	     const struct ifnet *out,
	     int offset,
	     const void *hdr,
	     u_int16_t datalen,
	     int *hotdrop)
{
	/* Stop iteration if it doesn't match */
	if (!m->u.kernel.match->match(skb, in, out, m->data,
				      offset, hdr, datalen, hotdrop))
		return 1;
	else
		return 0;
}

static inline struct ipt_entry *
get_entry(void *base, unsigned int offset)
{
	return (struct ipt_entry *)(PTROUT(base) + offset);
}

/* Returns whether matches rule or not. */
static inline int
ip_packet_match(const struct ip *ip,
		const char *indev,
		const char *outdev,
		const struct ipt_ip *ipinfo,
		int isfrag)
{
	size_t i;
	unsigned long ret;

#define FWINV(bool,invflg) ((bool) ^ !!(ipinfo->invflags & invflg))
	/* Matching source and destination address */
	if (FWINV((ip->ip_src.s_addr&ipinfo->smsk.s_addr) != ipinfo->src.s_addr,
		  IPT_INV_SRCIP)
	    || FWINV((ip->ip_dst.s_addr&ipinfo->dmsk.s_addr) != ipinfo->dst.s_addr,
		     IPT_INV_DSTIP)) {
		NFDEBUG("Source or dest mismatch.\n");
		return 0;
	}

	/* Look for ifname matches; this should unroll nicely. */
	/* Matching net device from */
	for (i = 0, ret = 0; i < IFNAMSIZ/sizeof(unsigned long); i++) {
		ret |= (((const unsigned long *)indev)[i]
			^ ((const unsigned long *)ipinfo->iniface)[i])
			& ((const unsigned long *)ipinfo->iniface_mask)[i];
	}

	if (FWINV(ret != 0, IPT_INV_VIA_IN)) {
		NFDEBUG("VIA in mismatch (%s vs %s).%s\n",
			indev, ipinfo->iniface,
			ipinfo->invflags&IPT_INV_VIA_IN ?" (INV)":"");
		return 0;
	}
	/* Matching net device to */
	for (i = 0, ret = 0; i < IFNAMSIZ/sizeof(unsigned long); i++) {
		ret |= (((const unsigned long *)outdev)[i]
			^ ((const unsigned long *)ipinfo->outiface)[i])
			& ((const unsigned long *)ipinfo->outiface_mask)[i];
	}

	if (FWINV(ret != 0, IPT_INV_VIA_OUT)) {
		NFDEBUG("VIA out mismatch (%s vs %s).%s\n",
			outdev, ipinfo->outiface,
			ipinfo->invflags&IPT_INV_VIA_OUT ?" (INV)":"");
		return 0;
	}

	/* Check specific protocol */
	/* Matching upper layer protocol */
	if (ipinfo->proto
	    && FWINV(ip->ip_p != ipinfo->proto, IPT_INV_PROTO)) {
		NFDEBUG("Packet protocol %hi does not match %hi.%s\n",
			ip->ip_p, ipinfo->proto,
			ipinfo->invflags&IPT_INV_PROTO ? " (INV)":"");
		return 0;
	}

	/* If we have a fragment rule but the packet is not a fragment
	 * then we return zero */
	if (FWINV((ipinfo->flags&IPT_F_FRAG) && !isfrag, IPT_INV_FRAG)) {
		NFDEBUG("Fragment rule but not fragment.%s\n",
			ipinfo->invflags & IPT_INV_FRAG ? " (INV)" : "");
		return 0;
	}

	return 1;
}

/* 
  * 	Returns one of the generic firewall policies, like NF_ACCEPT
  *	it is the main routine of ipt_hook()
  *   Note:ip->ip_len,ip->ip_off fields have been changed to
  *	host representation before run pfil hooks,see ip_input().
  *	After all,FreeBSD is __NOT__ Linux:)
  */
static unsigned int
ipt_do_table(struct mbuf **pskb,
	     unsigned int hook,
	     const struct ifnet *in,
	     const struct ifnet *out,
	     struct ipt_table *table,
	     void *userdata)
{
	static const char nulldevname[IFNAMSIZ] = { 0 };
	u_int16_t offset;
	struct ip *ip;
	void *protohdr;
	u_int16_t datalen;
	int hotdrop = 0;
	/* Initializing verdict to NF_DROP keeps gcc happy. */
	unsigned int verdict = NF_DROP;
	const char *indev, *outdev;
	void *table_base;
	struct ipt_entry *e, *back;

	/* Initialization */
	//KASSERT(hook == PFIL_IN, ("ipt_do_table() wrong direction!"));
	/* Get the ip protocol head */
	ip = mtod(*pskb, struct ip *);
	/* Get upper layer protocol head */
	protohdr = (u_int32_t *)ip + ip->ip_hl;
	datalen = (*pskb)->m_len - ip->ip_hl * 4;
	indev = in ? in->if_xname : nulldevname;
	outdev = out ? out->if_xname : nulldevname;
	/* We handle fragments by dealing with the first fragment as
	 * if it was a normal packet.  All other fragments are treated
	 * normally, except that they will NEVER match rules that ask
	 * things we don't know, ie. tcp syn flag or ports).  If the
	 * rule is also a fragment-specific rule, non-fragments won't
	 * match it. */
	offset = ip->ip_off & IP_OFFMASK;
	read_lock_bh(&table->lock);
	IP_NF_ASSERT(table->valid_hooks & (1 << hook));
	table_base = (void *)(table->private->entries
		+ TABLE_OFFSET(table->private,
			       cpu_number_map(smp_processor_id())));
	e = get_entry(table_base, table->private->hook_entry[hook]);
	
	#ifdef CONFIG_NETFILTER_DEBUG
	/* Check noone else using our table */
	if (((struct ipt_entry *)table_base)->comefrom != 0xdead57ac
	    && ((struct ipt_entry *)table_base)->comefrom != 0xeeeeeeec) {
		NFDEBUG("ASSERT: CPU #%u, %s comefrom(%p) = %X\n",
		       smp_processor_id(),
		       table->name,
		       &((struct ipt_entry *)table_base)->comefrom,
		       ((struct ipt_entry *)table_base)->comefrom);
	}
	((struct ipt_entry *)table_base)->comefrom = 0x57acc001;
	#endif

	/* For return from builtin chain */
	back = get_entry(table_base, table->private->underflow[hook]);

	do {
		IP_NF_ASSERT(e);
		IP_NF_ASSERT(back);
		//(*pskb)->nfcache |= e->nfcache;
		
		/* First,matching the general ip protocol head rules */
		if (ip_packet_match(ip, indev, outdev, &e->ip, offset)) {
			struct ipt_entry_target *t;

			/* Second,matching the matches */
			if (IPT_MATCH_ITERATE(e, do_match,
					      *pskb, in, out,
					      offset, protohdr,
					      datalen, &hotdrop) != 0)
				goto no_match;

			ADD_COUNTER(e->counters, ntohs(ip->ip_len), 1);

			t = ipt_get_target(e);
			IP_NF_ASSERT(t->u.kernel.target);
			/* Standard target? */
			if (!t->u.kernel.target->target) {
				int v;

				v = ((struct ipt_standard_target *)t)->verdict;
				if (v < 0) {
					/* Pop from stack? */
					if (v != IPT_RETURN) {
						verdict = (unsigned)(-v) - 1;
						break;
					}
					e = back;
					back = get_entry(table_base,
							 back->comefrom);
					continue;
				}
				if (PTROUT(table_base) + v
				    != PTROUT(e) + e->next_offset) {
					/* Save old back ptr in next entry */
					struct ipt_entry *next
						= (struct ipt_entry *)(PTROUT(e) + e->next_offset);
					next->comefrom
						= PTROUT(back) - PTROUT(table_base);
					/* set back pointer to next entry */
					back = next;
				}

				e = get_entry(table_base, v);
			} else {
				/* Targets which reenter must return
                                   abs. verdicts */
				#ifdef CONFIG_NETFILTER_DEBUG
				((struct ipt_entry *)table_base)->comefrom
					= 0xeeeeeeec;
				#endif
				/* Get the verdict to the packet */
				verdict = t->u.kernel.target->target(pskb,
								     hook,
								     in, out,
								     t->data,
								     userdata);

				#ifdef CONFIG_NETFILTER_DEBUG
				if (((struct ipt_entry *)table_base)->comefrom
				    != 0xeeeeeeec
				    && verdict == IPT_CONTINUE) {
					printf("Target %s reentered!\n",
					       t->u.kernel.target->name);
					verdict = NF_DROP;
				}
				((struct ipt_entry *)table_base)->comefrom
					= 0x57acc001;
				#endif
				/* Target might have changed stuff. */
				ip = mtod(*pskb, struct ip *);
				protohdr = (u_int32_t *)ip + ip->ip_hl;
				datalen = (*pskb)->m_len - ip->ip_hl * 4;

				if (verdict == IPT_CONTINUE)
					e = (struct ipt_entry *)(PTROUT(e) + e->next_offset);
				else
					/* Verdict */
					break;
			}
		} else {

		no_match:
			e = (struct ipt_entry *)(PTROUT(e) + e->next_offset);
		}
	} while (!hotdrop);

	#ifdef CONFIG_NETFILTER_DEBUG
	((struct ipt_entry *)table_base)->comefrom = 0xdead57ac;
	#endif
	read_unlock_bh(&table->lock);

	#ifdef DEBUG_ALLOW_ALL
	return NF_ACCEPT;
	#else
	if (hotdrop)
		return NF_DROP;
	else return verdict;
#endif
}

/* If it succeeds, returns element and locks mutex */
static inline void *
find_inlist_lock_noload(struct list_head *head,
			const char *name,
			int *error,
			struct mtx *mutex)
{
	void *ret;

#if 0
	printf("find_inlist: searching for `%s' in %s.\n",
		 name, head == &ipt_target ? "ipt_target"
		 : head == &ipt_match ? "ipt_match"
		 : head == &ipt_tables ? "ipt_tables" : "UNKNOWN");
#endif

	*error = down_interruptible(mutex);
	if (*error != 0)
		return NULL;

	ret = list_named_find(head, name);
	if (!ret) {
		*error = ENOENT;
		up(mutex);
	}
	return ret;
}

static void *
find_inlist_lock(struct list_head *head,
		 const char *name,
		 const char *prefix,
		 int *error,
		 struct mtx *mutex)
{
	void *ret;

	ret = find_inlist_lock_noload(head, name, error, mutex);
	if (!ret) {
		char modulename[IPT_FUNCTION_MAXNAMELEN + strlen(prefix) + 1];
		strcpy(modulename, prefix);
		strcat(modulename, name);
		log(LOG_ERR, "Netfilter/Iptables Error:no module '%s' support",modulename);
	}

	return ret;
}

static inline struct ipt_table *
find_table_lock(const char *name, int *error, struct mtx *mutex)
{
	return find_inlist_lock(&ipt_tables, name, "iptable_", error, mutex);
}

static inline struct ipt_match *
find_match_lock(const char *name, int *error, struct mtx *mutex)
{
	return find_inlist_lock(&ipt_match, name, "ipt_", error, mutex);
}

static inline struct ipt_target *
find_target_lock(const char *name, int *error, struct mtx *mutex)
{
	return find_inlist_lock(&ipt_target, name, "ipt_", error, mutex);
}


/* 
  *	Return 1:
  *	   1. Port is in [min...max] and invert not set
  *	   2.	Port is not in [min...max] and invert set
  */
static inline int
port_match(u_int16_t min, u_int16_t max, u_int16_t port, int invert)
{
	int ret;

	ret = (port >= min && port <= max) ^ invert;
	return ret;
}

/*
  *	Matching TCP option
  *	Return true or false: return FALSE and set *hotdrop = 1 to
  *	force immediate packet drop.
  *		@option:TCP option to be matched
  *		@tcp:TCP protocol head
  *		@datalen:header length
  *		@invert:invert flag
  *		@hotdrop:return value to indicate droping it immediately or not
  */
static int
tcp_find_option(u_int8_t option,
		const struct tcphdr *tcp,
		u_int16_t datalen,
		int invert,
		int *hotdrop)
{
	unsigned int i = sizeof(struct tcphdr);
	const u_int8_t *opt = (const u_int8_t *)tcp;

	NFDEBUG("tcp_match: finding option\n");
	
	/* If we don't have the whole header, drop packet. */
	if (tcp->th_off * 4 > datalen) {
		*hotdrop = 1;
		return 0;
	}

	/*
	  *	According to TCP protocol,the first byte of option is 'kind' field,
	  * 	if kind>1,the second byte is the length of the option
	  *	kind:
	  *		0:option end,total 1 byte
	  *		1:NULL operation,total 1 byte
	  */
	while (i < tcp->th_off * 4) {
		if (opt[i] == option) return !invert;
		if (opt[i] < 2) i++;
		else i += opt[i+1]?:1;
	}

	/* Goes here means option not matching */
	return invert;
}

/*
  *	Matching TCP rule
  *	Return true or false: return FALSE and set *hotdrop = 1 to
  *	force immediate packet drop.
  *		@skb:the packet buffer
  *		@in:net device the packet incoming from
  *		@out:net device the packet outgoing to
  *		@matchinfo:TCP matching stuff
  *		@offset:fragment flag,0 stands for not a fragment
  *		@hdr:the protocol head,here is struct tcphdr
  *		@datalen:header length
  *		@hotdrop:return value to indicate droping it immediately or not
  *
  */
static int
tcp_match(const struct mbuf *skb,
	  const struct ifnet *in,
	  const struct ifnet *out,
	  const void *matchinfo,
	  int offset,
	  const void *hdr,
	  u_int16_t datalen,
	  int *hotdrop)
{
	const struct tcphdr *tcp = hdr;
	const struct ipt_tcp *tcpinfo = matchinfo;
	
	/* To quote Alan:

	   Don't allow a fragment of TCP 8 bytes in. Nobody normal
	   causes this. Its a cracker trying to break in by doing a
	   flag overwrite to pass the direction checks.
	*/

	if (offset == 1) {
		NFDEBUG("Dropping evil TCP offset=1 frag.\n");
		*hotdrop = 1;
		return 0;
	} else if (offset == 0 && datalen < sizeof(struct tcphdr)) {
		/* We've been asked to examine this packet, and we
		   can't.  Hence, no choice but to drop. */
		NFDEBUG("Dropping evil TCP offset=0 tinygram.\n");
		*hotdrop = 1;
		return 0;
	}
	/* FIXME: Try tcp doff >> packet len against various stacks --RR */

#define FWINVTCP(bool,invflg) ((bool) ^ !!(tcpinfo->invflags & invflg))

	/* Must not be a fragment. */
	return !offset
	 	/* Matching TCP source port */
		&& port_match(tcpinfo->spts[0], tcpinfo->spts[1],
			      ntohs(tcp->th_sport),
			      !!(tcpinfo->invflags & IPT_TCP_INV_SRCPT))
		/* Matching TCP destination port */
		&& port_match(tcpinfo->dpts[0], tcpinfo->dpts[1],
			      ntohs(tcp->th_dport),
			      !!(tcpinfo->invflags & IPT_TCP_INV_DSTPT))
		/* Matching TCP flags */
		&& FWINVTCP((((const unsigned char *)tcp)[13]
			     & tcpinfo->flg_mask)
			    == tcpinfo->flg_cmp,
			    IPT_TCP_INV_FLAGS)
		/* Matching TCP option */
		&& (!tcpinfo->option
		    || tcp_find_option(tcpinfo->option, tcp, datalen,
				       tcpinfo->invflags
				       & IPT_TCP_INV_OPTION,
				       hotdrop));
}

#define IPT_ALIGN(s) (((s) + (__alignof__(struct ipt_entry)-1)) & ~(__alignof__(struct ipt_entry)-1))

/* Called when user tries to insert an entry of this type. */
static int
tcp_checkentry(const char *tablename,
	       const struct ipt_ip *ip,
	       void *matchinfo,
	       unsigned int matchsize,
	       unsigned int hook_mask)
{
	const struct ipt_tcp *tcpinfo = matchinfo;

	/* Must specify proto == TCP, and no unknown invflags */
	return ip->proto == IPPROTO_TCP
		&& !(ip->invflags & IPT_INV_PROTO)
		&& matchsize == IPT_ALIGN(sizeof(struct ipt_tcp))
		&& !(tcpinfo->invflags & ~IPT_TCP_INV_MASK);
}

/*
  *	Matching UDP rule
  *	Return true or false: return FALSE and set *hotdrop = 1 to
  *	force immediate packet drop.
  *		@skb:the packet buffer
  *		@in:net device the packet incoming from
  *		@out:net device the packet outgoing to
  *		@matchinfo:UDP matching stuff
  *		@offset:fragment flag,0 stands for not a fragment
  *		@hdr:the protocol head,here is struct udphdr
  *		@datalen:header length
  *		@hotdrop:return value to indicate droping it immediately or not
  *
  */
static int
udp_match(const struct mbuf *skb,
	  const struct ifnet *in,
	  const struct ifnet *out,
	  const void *matchinfo,
	  int offset,
	  const void *hdr,
	  u_int16_t datalen,
	  int *hotdrop)
{
	const struct udphdr *udp = hdr;
	const struct ipt_udp *udpinfo = matchinfo;

	if (offset == 0 && datalen < sizeof(struct udphdr)) {
		/* We've been asked to examine this packet, and we
		   can't.  Hence, no choice but to drop. */
		NFDEBUG("Dropping evil UDP tinygram.\n");
		*hotdrop = 1;
		return 0;
	}

	/* Must not be a fragment. */
	return !offset
		/* Matching UDP source port */
		&& port_match(udpinfo->spts[0], udpinfo->spts[1],
			      ntohs(udp->uh_sport),
			      !!(udpinfo->invflags & IPT_UDP_INV_SRCPT))
		/* Matching UDP destination port */
		&& port_match(udpinfo->dpts[0], udpinfo->dpts[1],
			      ntohs(udp->uh_dport),
			      !!(udpinfo->invflags & IPT_UDP_INV_DSTPT));
}

/* Called when user tries to insert an entry of this type. */
static int
udp_checkentry(const char *tablename,
	       const struct ipt_ip *ip,
	       void *matchinfo,
	       unsigned int matchinfosize,
	       unsigned int hook_mask)
{
	const struct ipt_udp *udpinfo = matchinfo;

	/* Must specify proto == UDP, and no unknown invflags */
	if (ip->proto != IPPROTO_UDP || (ip->invflags & IPT_INV_PROTO)) {
		NFDEBUG("ipt_udp: Protocol %u != %u\n", ip->proto,
			 IPPROTO_UDP);
		return 0;
	}
	if (matchinfosize != IPT_ALIGN(sizeof(struct ipt_udp))) {
		NFDEBUG("ipt_udp: matchsize %u != %u\n",
			 matchinfosize, IPT_ALIGN(sizeof(struct ipt_udp)));
		return 0;
	}
	if (udpinfo->invflags & ~IPT_UDP_INV_MASK) {
		NFDEBUG("ipt_udp: unknown flags %X\n",
			 udpinfo->invflags);
		return 0;
	}

	return 1;
}

/*
  *	Return 1:
  *		1.Type equal and code is in [min_code...max_code] and invert not set
  *		2.Type not equal or code is not in [min_code...max_code] and invert set
  */
static inline int
icmp_type_code_match(u_int8_t test_type, u_int8_t min_code, u_int8_t max_code,
		     u_int8_t type, u_int8_t code,
		     int invert)
{
	return (type == test_type && code >= min_code && code <= max_code)
		^ invert;
}


/*
  *	Matching ICMP rule
  *	Return true or false: return FALSE and set *hotdrop = 1 to
  *	force immediate packet drop.
  *		@skb:the packet buffer
  *		@in:net device the packet incoming from
  *		@out:net device the packet outgoing to
  *		@matchinfo:ICMP matching stuff
  *		@offset:fragment flag,0 stands for not a fragment
  *		@hdr:the protocol head,here is struct icmp
  *		@datalen:header length
  *		@hotdrop:return value to indicate droping it immediately or not
  *
  */
static int
icmp_match(const struct mbuf *skb,
	   const struct ifnet *in,
	   const struct ifnet *out,
	   const void *matchinfo,
	   int offset,
	   const void *hdr,
	   u_int16_t datalen,
	   int *hotdrop)
{
	const struct icmp *icmp = hdr;
	const struct ipt_icmp *icmpinfo = matchinfo;

	if (offset == 0 && datalen < 2) {
		/* We've been asked to examine this packet, and we
		   can't.  Hence, no choice but to drop. */
		NFDEBUG("Dropping evil ICMP tinygram.\n");
		*hotdrop = 1;
		return 0;
	}

	/* Must not be a fragment. */
	return !offset
		&& icmp_type_code_match(icmpinfo->type,
					icmpinfo->code[0],
					icmpinfo->code[1],
					icmp->icmp_type, icmp->icmp_code,
					!!(icmpinfo->invflags&IPT_ICMP_INV));
}

/* Called when user tries to insert an entry of this type. */
static int
icmp_checkentry(const char *tablename,
	   const struct ipt_ip *ip,
	   void *matchinfo,
	   unsigned int matchsize,
	   unsigned int hook_mask)
{
	const struct ipt_icmp *icmpinfo = matchinfo;

	/* Must specify proto == ICMP, and no unknown invflags */
	return ip->proto == IPPROTO_ICMP
		&& !(ip->invflags & IPT_INV_PROTO)
		&& matchsize == IPT_ALIGN(sizeof(struct ipt_icmp))
		&& !(icmpinfo->invflags & ~IPT_ICMP_INV);
}

static inline int
ip_checkentry(struct ipt_ip *ip)
{
	if (ip->flags & ~IPT_F_MASK) {
		NFDEBUG("Unknown flag bits set: %08X\n",
			 ip->flags & ~IPT_F_MASK);
		return 0;
	}
	if (ip->invflags & ~IPT_INV_MASK) {
		NFDEBUG("Unknown invflag bits set: %08X\n",
			 ip->invflags & ~IPT_INV_MASK);
		return 0;
	}
	
	return 1;
}

#define 	net_msg_cost 	5*hz
#define	net_msg_burst	10*5*hz

/* 
 * This enforces a rate limit: not more than one kernel message
 * every 5secs to make a denial-of-service attack impossible.
 *
 * All warning printf()s should be guarded by this function. 
 */ 
int net_ratelimit(void)
{
	static struct mtx ratelimit_lock;
	static unsigned long toks;
	static unsigned long last_msg; 
	static int missed;
	unsigned long now = ticks;

	if (!mtx_initialized(&ratelimit_lock)) {
		mtx_init(&ratelimit_lock,"ratelimit lock",NULL,MTX_DEF);
		toks=10*5*hz;
	}
	mtx_lock(&ratelimit_lock);
	toks += now - last_msg;
	last_msg = now;
	if (toks > net_msg_burst)
		toks = net_msg_burst;
	if (toks >= net_msg_cost) {
		int lost = missed;
		missed = 0;
		toks -= net_msg_cost;
		mtx_unlock(&ratelimit_lock);
		if (lost)
			log(LOG_INFO, "NET: %d messages suppressed.\n", lost);
		return 1;
	}
	missed++;
	mtx_unlock(&ratelimit_lock);
	return 0;
}

static unsigned int
ipt_error(struct mbuf **pskb,
	  unsigned int hooknum,
	  const struct ifnet *in,
	  const struct ifnet *out,
	  const void *targinfo,
	  void *userinfo)
{
	
	if (net_ratelimit())
		log(LOG_ERR, "ip_tables: error: `%s'\n", (const char *)targinfo);

	return 0;
}

static int do_ipt_set_ctl(struct socket *,int,void *,unsigned int);
static int do_ipt_get_ctl(struct socket *, int, void *, int *);

static struct ipt_target ipt_standard_target
= { { NULL, NULL }, IPT_STANDARD_TARGET, NULL, NULL, NULL };

static struct ipt_target ipt_error_target
= { { NULL, NULL }, IPT_ERROR_TARGET, ipt_error, NULL, NULL };

static struct nf_sockopt_ops ipt_sockopts
= { { NULL, NULL }, PF_INET, IPT_BASE_CTL, IPT_SO_SET_MAX+1, do_ipt_set_ctl,
    IPT_BASE_CTL, IPT_SO_GET_MAX+1, do_ipt_get_ctl, 0, { NULL, 0 }, NULL  };

static struct ipt_match tcp_matchstruct
= { { NULL, NULL }, "tcp", &tcp_match, &tcp_checkentry, NULL };
static struct ipt_match udp_matchstruct
= { { NULL, NULL }, "udp", &udp_match, &udp_checkentry, NULL };
static struct ipt_match icmp_matchstruct
= { { NULL, NULL }, "icmp", &icmp_match, &icmp_checkentry, NULL };

/* All zeroes == unconditional rule. */
static inline int
unconditional(const struct ipt_ip *ip)
{
	unsigned int i;

	for (i = 0; i < sizeof(*ip)/sizeof(u_int32_t); i++)
		if (((const u_int32_t *)ip)[i])
			return 0;

	return 1;
}

/* 
  * Figures out from what hook each rule can be called: returns 0 if
  * there are loops.  Puts hook bitmask in comefrom.
  */
static int
mark_source_chains(struct ipt_table_info *newinfo, unsigned int valid_hooks)
{
	unsigned int hook;

	/* No recursion; use packet counter to save back ptrs (reset
	   to 0 as we leave), and comefrom to save source hook bitmask */
	for (hook = 0; hook < NF_IP_NUMHOOKS; hook++) {
		unsigned int pos = newinfo->hook_entry[hook];
		struct ipt_entry *e
			= (struct ipt_entry *)(newinfo->entries + pos);

		if (!(valid_hooks & (1 << hook)))
			continue;

		/* Set initial back pointer. */
		e->counters.pcnt = pos;

		for (;;) {
			struct ipt_standard_target *t
				= (void *)ipt_get_target(e);

			if (e->comefrom & (1 << NF_IP_NUMHOOKS)) {
				NFDEBUG("iptables: loop hook %u pos %u %08X.\n",
				       hook, pos, e->comefrom);
				return 0;
			}
			e->comefrom
				|= ((1 << hook) | (1 << NF_IP_NUMHOOKS));

			/* Unconditional return/END. */
			if (e->target_offset == sizeof(struct ipt_entry)  /* No matches */
			    && (strcmp(t->target.u.user.name,    /* Standard target:'' */
				       IPT_STANDARD_TARGET) == 0)
			    && t->verdict < 0  /* IPT_VONTINUE means continue to next */
			    && unconditional(&e->ip)) {  /* Check the struct ipt_ip whether be all zero */
				unsigned int oldpos, size;

				/* Return: backtrack through the last
				   big jump. */
				do {
					e->comefrom ^= (1<<NF_IP_NUMHOOKS);
					#ifdef DEBUG_IP_FIREWALL_USER
					if (e->comefrom
					    & (1 << NF_IP_NUMHOOKS)) {
						NFDEBUG("Back unset "
							 "on hook %u "
							 "rule %u\n",
							 hook, pos);
					}
					#endif
					oldpos = pos;
					pos = e->counters.pcnt;
					e->counters.pcnt = 0;

					/* We're at the start. */
					if (pos == oldpos)
						goto next;

					e = (struct ipt_entry *)
						(newinfo->entries + pos);
				} while (oldpos == pos + e->next_offset);

				/* Move along one */
				size = e->next_offset;
				e = (struct ipt_entry *)
					(newinfo->entries + pos + size);
				e->counters.pcnt = pos;
				pos += size;
			} else {
				int newpos = t->verdict;

				if (strcmp(t->target.u.user.name,
					   IPT_STANDARD_TARGET) == 0
				    && newpos >= 0) {
					/* This a jump; chase it. */
					NFDEBUG("Jump rule %u -> %u\n",
						 pos, newpos);
				} else {
					/* ... this is a fallthru */
					newpos = pos + e->next_offset;
				}
				e = (struct ipt_entry *)
					(newinfo->entries + newpos);
				e->counters.pcnt = pos;
				pos = newpos;
			}
		}
		next:
		NFDEBUG("Finished chain %u\n", hook);
	}
	return 1;
}

static inline int
cleanup_match(struct ipt_entry_match *m, unsigned int *i)
{
	if (i && (*i)-- == 0)
		return 1;

	if (m->u.kernel.match->destroy)
		m->u.kernel.match->destroy(m->data,
					   m->u.match_size - sizeof(*m));

	return 0;
}

static inline int
standard_check(struct ipt_entry_target *t,
	       unsigned int max_offset)
{
	struct ipt_standard_target *targ = (struct ipt_standard_target *)t;

	/* Check standard info. */
	if (t->u.target_size
	    != IPT_ALIGN(sizeof(struct ipt_standard_target))) {
		NFDEBUG("standard_check: target size %u != %u\n",
			 t->u.target_size,
			 IPT_ALIGN(sizeof(struct ipt_standard_target)));
		return 0;
	}

	if (targ->verdict >= 0
	    && targ->verdict > max_offset - sizeof(struct ipt_entry)) {
		NFDEBUG("ipt_standard_check: bad verdict (%i)\n",
			 targ->verdict);
		return 0;
	}

	if (targ->verdict < -NF_MAX_VERDICT - 1) {
		NFDEBUG("ipt_standard_check: bad negative verdict (%i)\n",
			 targ->verdict);
		return 0;
	}
	return 1;
}

static inline int
check_match(struct ipt_entry_match *m,
	    const char *name,
	    const struct ipt_ip *ip,
	    unsigned int hookmask,
	    unsigned int *i)
{
	int ret;
	struct ipt_match *match;

	match = find_match_lock(m->u.user.name, &ret, &ipt_mutex);
	if (!match) {
		NFDEBUG("check_match: `%s' not found\n", m->u.user.name);
		return ret;
	}
	m->u.kernel.match = match;
	up(&ipt_mutex);

	if (m->u.kernel.match->checkentry
	    && !m->u.kernel.match->checkentry(name, ip, m->data,
					      m->u.match_size - sizeof(*m),
					      hookmask)) {
		NFDEBUG("ip_tables: check failed for `%s'.\n",
			 m->u.kernel.match->name);
		return EINVAL;
	}

	(*i)++;
	return 0;
}

static inline int
check_entry(struct ipt_entry *e, const char *name, unsigned int size,
	    unsigned int *i)
{
	struct ipt_entry_target *t;
	struct ipt_target *target;
	int ret;
	unsigned int j;

	if (!ip_checkentry(&e->ip)) {
		NFDEBUG("ip_tables: ip check failed %p %s.\n", e, name);
		return EINVAL;
	}

	j = 0;
	ret = IPT_MATCH_ITERATE(e, check_match, name, &e->ip, e->comefrom, &j);
	if (ret != 0)
		goto cleanup_matches;

	t = ipt_get_target(e);
	target = find_target_lock(t->u.user.name, &ret, &ipt_mutex);
	if (!target) {
		NFDEBUG("check_entry: `%s' not found\n", t->u.user.name);
		goto cleanup_matches;
	}
	t->u.kernel.target = target;
	up(&ipt_mutex);

	if (t->u.kernel.target == &ipt_standard_target) {
		if (!standard_check(t, size)) {
			ret = EINVAL;
			goto cleanup_matches;
		}
	} else if (t->u.kernel.target->checkentry
		   && !t->u.kernel.target->checkentry(name, e, t->data,
						      t->u.target_size
						      - sizeof(*t),
						      e->comefrom)) {
		NFDEBUG("ip_tables: check failed for `%s'.\n",
			 t->u.kernel.target->name);
		ret = EINVAL;
		goto cleanup_matches;
	}

	(*i)++;
	return 0;

 cleanup_matches:
	IPT_MATCH_ITERATE(e, cleanup_match, &j);
	return ret;
}

static inline int
check_entry_size_and_hooks(struct ipt_entry *e,
			   struct ipt_table_info *newinfo,
			   unsigned char *base,
			   unsigned char *limit,
			   const unsigned int *hook_entries,
			   const unsigned int *underflows,
			   unsigned int *i)
{
	unsigned int h;

	if ((unsigned long)e % __alignof__(struct ipt_entry) != 0
	    || (unsigned char *)e + sizeof(struct ipt_entry) >= limit) {
		NFDEBUG("Bad offset %p\n", e);
		return EINVAL;
	}

	/* The min size of rule:size of ipt_entry + ipt_entry_target,no matches */
	if (e->next_offset
	    < sizeof(struct ipt_entry) + sizeof(struct ipt_entry_target)) {
		NFDEBUG("checking: element %p size %u\n",
			 e, e->next_offset);
		return EINVAL;
	}

	/* Check hooks & underflows */
	for (h = 0; h < NF_IP_NUMHOOKS; h++) {
		if ((unsigned char *)e - base == hook_entries[h])
			newinfo->hook_entry[h] = hook_entries[h];
		if ((unsigned char *)e - base == underflows[h])
			newinfo->underflow[h] = underflows[h];
	}

	/* FIXME: underflows must be unconditional, standard verdicts
           < 0 (not IPT_RETURN). --RR */

	/* Clear counters and comefrom */
	e->counters = ((struct ipt_counters) { 0, 0 });
	e->comefrom = 0;

	(*i)++;
	return 0;
}

static inline int
cleanup_entry(struct ipt_entry *e, unsigned int *i)
{
	struct ipt_entry_target *t;

	if (i && (*i)-- == 0)
		return 1;

	/* Cleanup all matches */
	IPT_MATCH_ITERATE(e, cleanup_match, NULL);
	t = ipt_get_target(e);
	if (t->u.kernel.target->destroy)
		t->u.kernel.target->destroy(t->data,
					    t->u.target_size - sizeof(*t));

	return 0;
}

/* Standard entry. */
struct ipt_standard
{
	struct ipt_entry entry;
	struct ipt_standard_target target;
};

/* 
  *	Checks and translates the user-supplied table segment (held in
  *	newinfo)
  *		@name:table name
  *		@valid_hooks:what hooks you will enter on
  *		@newinfo:newly malloced ipt_table_info structure
  *		@size:total size of new entries
  *		@number:number of entries
  *		@hook_entries:hook entry points
  *		@underflows:underflow points
  */
static int
translate_table(const char *name,
		unsigned int valid_hooks,
		struct ipt_table_info *newinfo,
		unsigned int size,
		unsigned int number,
		const unsigned int *hook_entries,
		const unsigned int *underflows)
{
	unsigned int i;
	int ret;

	newinfo->size = size;
	newinfo->number = number;

	/* Init all hooks to impossible value. */
	for (i = 0; i < NF_IP_NUMHOOKS; i++) {
		newinfo->hook_entry[i] = 0xFFFFFFFF;
		newinfo->underflow[i] = 0xFFFFFFFF;
	}

	NFDEBUG("translate_table: size %u\n", newinfo->size);
	i = 0;
	/* Walk through entries, checking offsets. */
	ret = IPT_ENTRY_ITERATE(newinfo->entries, newinfo->size,
				check_entry_size_and_hooks,
				newinfo,
				newinfo->entries,
				newinfo->entries + size,
				hook_entries, underflows, &i);
	if (ret != 0)
		return ret;

	if (i != number) {
		NFDEBUG("translate_table: %u not %u entries\n",
			 i, number);
		return EINVAL;
	}

	/* Check hooks all assigned */
	for (i = 0; i < NF_IP_NUMHOOKS; i++) {
		/* Only hooks which are valid */
		if (!(valid_hooks & (1 << i)))
			continue;
		if (newinfo->hook_entry[i] == 0xFFFFFFFF) {
			NFDEBUG("Invalid hook entry %u %u\n",
				 i, hook_entries[i]);
			return EINVAL;
		}
		if (newinfo->underflow[i] == 0xFFFFFFFF) {
			NFDEBUG("Invalid underflow %u %u\n",
				 i, underflows[i]);
			return EINVAL;
		}
	}

	if (!mark_source_chains(newinfo, valid_hooks))
		return ELOOP;

	/* Finally, each sanity check must pass */
	i = 0;
	ret = IPT_ENTRY_ITERATE(newinfo->entries, newinfo->size,
				check_entry, name, size, &i);

	if (ret != 0) {
		IPT_ENTRY_ITERATE(newinfo->entries, newinfo->size,
				  cleanup_entry, &i);
		return ret;
	}

	/* And one copy for every other CPU */
	for (i = 1; i < smp_num_cpus; i++) {
		memcpy(newinfo->entries + SMP_ALIGN(newinfo->size)*i,
		       newinfo->entries,
		       SMP_ALIGN(newinfo->size));
	}

	return ret;
}

static struct ipt_table_info *
replace_table(struct ipt_table *table,
	      unsigned int num_counters,
	      struct ipt_table_info *newinfo,
	      int *error)
{
	struct ipt_table_info *oldinfo;

	#ifdef CONFIG_NETFILTER_DEBUG
	{
		struct ipt_entry *table_base;
		unsigned int i;

		for (i = 0; i < smp_num_cpus; i++) {
			table_base =
				(void *)newinfo->entries
				+ TABLE_OFFSET(newinfo, i);

			table_base->comefrom = 0xdead57ac;
		}
	}
	#endif

	/* Do the substitution. */
	write_lock_bh(&table->lock);
	/* Check inside lock: is the old number correct? */
	if (num_counters != table->private->number) {
		NFDEBUG("num_counters != table->private->number (%u/%u)\n",
			 num_counters, table->private->number);
		write_unlock_bh(&table->lock);
		*error = EAGAIN;
		return NULL;
	}
	oldinfo = table->private;
	table->private = newinfo;
	newinfo->initial_entries = oldinfo->initial_entries;
	write_unlock_bh(&table->lock);

	return oldinfo;
}

/* Do exclusive ranges overlap? */
static inline int overlap(int min1, int max1, int min2, int max2)
{
	return max1 > min2 && min1 < max2;
}

/* Gets counters. */
static inline int
add_entry_to_counter(const struct ipt_entry *e,
		     struct ipt_counters total[],
		     unsigned int *i)
{
	ADD_COUNTER(total[*i], e->counters.bcnt, e->counters.pcnt);

	(*i)++;
	return 0;
}

static void
get_counters(const struct ipt_table_info *t,
	     struct ipt_counters counters[])
{
	unsigned int cpu;
	unsigned int i;

	for (cpu = 0; cpu < smp_num_cpus; cpu++) {
		i = 0;
		IPT_ENTRY_ITERATE(t->entries + TABLE_OFFSET(t, cpu),
				  t->size,
				  add_entry_to_counter,
				  counters,
				  &i);
	}
}

static int
copy_entries_to_user(unsigned int total_size,
		     struct ipt_table *table,
		     void *userptr)
{
	unsigned int off, num, countersize;
	struct ipt_entry *e;
	struct ipt_counters *counters;
	int ret = 0;

	/* We need atomic snapshot of counters: rest doesn't change
	   (other than comefrom, which userspace doesn't care
	   about). */
	countersize = sizeof(struct ipt_counters) * table->private->number;
	counters=malloc(countersize,M_IPTBUF,M_WAITOK|M_ZERO);
	if (counters == NULL)
		return ENOMEM;

	/* First, sum counters... */
	memset(counters, 0, countersize);
	write_lock_bh(&table->lock);
	get_counters(table->private, counters);
	write_unlock_bh(&table->lock);

	/* ... then copy entire thing from CPU 0... */
	if (copy_to_user(userptr, table->private->entries, total_size) != 0) {
		ret = EFAULT;
		goto free_counters;
	}

	/* FIXME: use iterator macros --RR */
	/* ... then go back and fix counters and names */
	for (off = 0, num = 0; off < total_size; off += e->next_offset, num++){
		unsigned int i;
		struct ipt_entry_match *m;
		struct ipt_entry_target *t;

		e = (struct ipt_entry *)(table->private->entries + off);
		if (copy_to_user((void *)(PTROUT(userptr) + off
				 + offsetof(struct ipt_entry, counters)),
				 &counters[num],
				 sizeof(counters[num])) != 0) {
			ret = EFAULT;
			goto free_counters;
		}

		for (i = sizeof(struct ipt_entry);
		     i < e->target_offset;
		     i += m->u.match_size) {
			m = (struct ipt_entry_match *)(PTROUT(e) + i);

			if (copy_to_user((void *)(PTROUT(userptr) + off + i
					 + offsetof(struct ipt_entry_match,
						    u.user.name)),
					 m->u.kernel.match->name,
					 strlen(m->u.kernel.match->name)+1)
			    != 0) {
				ret = EFAULT;
				goto free_counters;
			}
		}

		t = ipt_get_target(e);
		if (copy_to_user((void *)(PTROUT(userptr) + off + e->target_offset
				 + offsetof(struct ipt_entry_target,
					    u.user.name)),
				 t->u.kernel.target->name,
				 strlen(t->u.kernel.target->name)+1) != 0) {
			ret = EFAULT;
			goto free_counters;
		}
	}

 free_counters:
	free(counters,M_IPTBUF);
	return ret;
}

static int
get_entries(const struct ipt_get_entries *entries,
	    struct ipt_get_entries *uptr)
{
	int ret;
	struct ipt_table *t;

	t = find_table_lock(entries->name, &ret, &ipt_mutex);
	if (t) {
		NFDEBUG("t->private->number = %u\n",
			 t->private->number);
		if (entries->size == t->private->size)
			ret = copy_entries_to_user(t->private->size,
						   t, uptr->entrytable);
		else {
			NFDEBUG("get_entries: I've got %u not %u!\n",
				 t->private->size,
				 entries->size);
			ret = EINVAL;
		}
		up(&ipt_mutex);
	} else
		NFDEBUG("get_entries: Can't find %s!\n",
			 entries->name);

	return ret;
}

static int
do_replace(void *user, unsigned int len)
{
	int ret;
	struct ipt_replace tmp;
	struct ipt_table *t;
	struct ipt_table_info *newinfo, *oldinfo;
	struct ipt_counters *counters;

	if (copy_from_user(&tmp, user, sizeof(tmp)) != 0)
		return EFAULT;

	/* Hack: Causes ipchains to give correct error msg --RR */
	if (len != sizeof(tmp) + tmp.size)
		return ENOPROTOOPT;

	/* Pedantry: prevent them from hitting BUG() in vmalloc.c --RR */
	/*
	if ((SMP_ALIGN(tmp.size) >> PAGE_SHIFT) + 2 > num_physpages)
		return -ENOMEM;*/

	newinfo = malloc(sizeof(struct ipt_table_info)
			  + SMP_ALIGN(tmp.size) * smp_num_cpus,M_IPTBUF,M_WAITOK|M_ZERO);
	if (!newinfo)
		return ENOMEM;

	if (copy_from_user(newinfo->entries, (void *)(PTROUT(user) + sizeof(tmp)),
			   tmp.size) != 0) {
		ret = EFAULT;
		goto free_newinfo;
	}

	counters = malloc(tmp.num_counters * sizeof(struct ipt_counters),M_IPTBUF,M_WAITOK|M_ZERO);
	if (!counters) {
		ret = ENOMEM;
		goto free_newinfo;
	}
	memset(counters, 0, tmp.num_counters * sizeof(struct ipt_counters));

	ret = translate_table(tmp.name, tmp.valid_hooks,
			      newinfo, tmp.size, tmp.num_entries,
			      tmp.hook_entry, tmp.underflow);
	if (ret != 0)
		goto free_newinfo_counters;

	NFDEBUG("ip_tables: Translated table\n");

	t = find_table_lock(tmp.name, &ret, &ipt_mutex);
	if (!t)
		goto free_newinfo_counters_untrans;

	/* You lied! */
	if (tmp.valid_hooks != t->valid_hooks) {
		NFDEBUG("Valid hook crap: %08X vs %08X\n",
			 tmp.valid_hooks, t->valid_hooks);
		ret = EINVAL;
		goto free_newinfo_counters_untrans_unlock;
	}

	oldinfo = replace_table(t, tmp.num_counters, newinfo, &ret);
	if (!oldinfo)
		goto free_newinfo_counters_untrans_unlock;

	/* Update module usage count based on number of rules */
	NFDEBUG("do_replace: oldnum=%u, initnum=%u, newnum=%u\n",
		oldinfo->number, oldinfo->initial_entries, newinfo->number);
	
	/* Get the old counters. */
	get_counters(oldinfo, counters);
	/* Decrease module usage counts and free resource */
	IPT_ENTRY_ITERATE(oldinfo->entries, oldinfo->size, cleanup_entry,NULL);
	free(oldinfo,M_IPTBUF);
	/* Silent error: too late now. */
	copy_to_user(tmp.counters, counters,
		     sizeof(struct ipt_counters) * tmp.num_counters);
	free(counters,M_IPTBUF);
	up(&ipt_mutex);
	return 0;

 free_newinfo_counters_untrans_unlock:
	up(&ipt_mutex);
 free_newinfo_counters_untrans:
	IPT_ENTRY_ITERATE(newinfo->entries, newinfo->size, cleanup_entry,NULL);
 free_newinfo_counters:
	free(counters,M_IPTBUF);
 free_newinfo:
	free(newinfo,M_IPTBUF);
	return ret;
}

/* We're lazy, and add to the first CPU; overflow works its fey magic
 * and everything is OK. */
static inline int
add_counter_to_entry(struct ipt_entry *e,
		     const struct ipt_counters addme[],
		     unsigned int *i)
{
#if 0
	duprintf("add_counter: Entry %u %lu/%lu + %lu/%lu\n",
		 *i,
		 (long unsigned int)e->counters.pcnt,
		 (long unsigned int)e->counters.bcnt,
		 (long unsigned int)addme[*i].pcnt,
		 (long unsigned int)addme[*i].bcnt);
#endif

	ADD_COUNTER(e->counters, addme[*i].bcnt, addme[*i].pcnt);

	(*i)++;
	return 0;
}

static int
do_add_counters(void *user, unsigned int len)
{
	unsigned int i;
	struct ipt_counters_info tmp, *paddc;
	struct ipt_table *t;
	int ret;

	if (copy_from_user(&tmp, user, sizeof(tmp)) != 0)
		return EFAULT;

	if (len != sizeof(tmp) + tmp.num_counters*sizeof(struct ipt_counters))
		return EINVAL;

	paddc = malloc(len,M_IPTBUF,M_WAITOK|M_ZERO);
	if (!paddc)
		return ENOMEM;

	if (copy_from_user(paddc, user, len) != 0) {
		ret = EFAULT;
		goto free;
	}

	t = find_table_lock(tmp.name, &ret, &ipt_mutex);
	if (!t)
		goto free;

	write_lock_bh(&t->lock);
	if (t->private->number != paddc->num_counters) {
		ret = EINVAL;
		goto unlock_up_free;
	}

	i = 0;
	IPT_ENTRY_ITERATE(t->private->entries,
			  t->private->size,
			  add_counter_to_entry,
			  paddc->counters,
			  &i);
 unlock_up_free:
	write_unlock_bh(&t->lock);
	up(&ipt_mutex);
 free:
	free(paddc,M_IPTBUF);

	return ret;
}

static int
do_ipt_set_ctl(struct socket *sk,int cmd, void *user, unsigned int len)
{
	int ret=0;

	if (suser(curthread))
		return EPERM;
	
	switch (cmd) {
	case IPT_SO_SET_REPLACE:
		ret = do_replace(user, len);
		break;

	case IPT_SO_SET_ADD_COUNTERS:
		ret = do_add_counters(user, len);
		break;

	default:
		log(LOG_INFO, "do_ipt_set_ctl:  unknown request %i\n", cmd);
		ret = EINVAL;
	}
	
	return ret;
}

static int
do_ipt_get_ctl(struct socket *sk, int cmd, void *user, int *len)
{
	int ret=0;

	if (suser(curthread))
		return EPERM;

	switch (cmd) { 
	case IPT_SO_GET_INFO: {
		char name[IPT_TABLE_MAXNAMELEN];
		struct ipt_table *t;

		if (*len != sizeof(struct ipt_getinfo)) {
			NFDEBUG("length %u != %u\n", *len,
				 sizeof(struct ipt_getinfo));
			ret = EINVAL;
			break;
		}

		if (copy_from_user(name, user, sizeof(name)) != 0) {
			ret = EFAULT;
			break;
		}
		name[IPT_TABLE_MAXNAMELEN-1] = '\0';
		t = find_table_lock(name, &ret, &ipt_mutex);
		if (t) {
			struct ipt_getinfo info;

			info.valid_hooks = t->valid_hooks;
			memcpy(info.hook_entry, t->private->hook_entry,
			       sizeof(info.hook_entry));
			memcpy(info.underflow, t->private->underflow,
			       sizeof(info.underflow));
			info.num_entries = t->private->number;
			info.size = t->private->size;
			strcpy(info.name, name);

			if (copy_to_user(user, &info, *len) != 0)
				ret = EFAULT;
			else
				ret = 0;

			up(&ipt_mutex);
		} else {
			ret=ENOENT;
		}
	}
	break; 
	
	case IPT_SO_GET_ENTRIES: {
		struct ipt_get_entries get;

		if (*len < sizeof(get)) {
			NFDEBUG("get_entries: %u < %u\n", *len, sizeof(get));
			ret = EINVAL;
		} else if (copy_from_user(&get, user, sizeof(get)) != 0) {
			ret = EFAULT;
		} else if (*len != sizeof(struct ipt_get_entries) + get.size) {
			NFDEBUG("get_entries: %u != %u\n", *len,
				 sizeof(struct ipt_get_entries) + get.size);
			ret = EINVAL;
		} else
			ret = get_entries(&get, user);
		break;
	}

	default:
		log(LOG_INFO, "do_ipt_get_ctl: unknown request %i\n", cmd);
		ret = EINVAL;
	}
	
	return ret;
}

/*
  *	Add a table
  *   default built-in table:filter,nat,mange
  */
int ipt_register_table(struct ipt_table *table)
{
	int ret;
	struct ipt_table_info *newinfo;  /* the table info supplied to add a table */
	static struct ipt_table_info bootstrap
		= { 0, 0, 0, { 0 }, { 0 }, { } };

	/* size = size of ipt_table_info + all entries */
	newinfo = malloc(sizeof(struct ipt_table_info)
			  + SMP_ALIGN(table->table->size) * smp_num_cpus,
			  M_IPTBUF,
			  M_WAITOK | M_ZERO);
	if (!newinfo) {
		ret = ENOMEM;
		return ret;
	}
	/* Copy the entries to table template */
	memcpy(newinfo->entries, table->table->entries, table->table->size);

	ret = translate_table(table->name, table->valid_hooks,
			      newinfo, table->table->size,
			      table->table->num_entries,
			      table->table->hook_entry,
			      table->table->underflow);
	
	if (ret != 0) {
		free(newinfo,M_IPTBUF);
		return ret;
	}

	ret = down_interruptible(&ipt_mutex);
	if (ret != 0) {
		free(newinfo,M_IPTBUF);
		return ret;
	}

	/* Don't autoload: we'd eat our tail... */
	if (list_named_find(&ipt_tables, table->name)) {
		ret = EEXIST;
		goto free_unlock;
	}

	/* Simplifies replace_table code. */
	table->private = &bootstrap;
	if (!replace_table(table, 0, newinfo, &ret))
		goto free_unlock;

	NFDEBUG("table->private->number = %u\n",
		 table->private->number);
	
	/* save number of initial entries */
	table->private->initial_entries = table->private->number;

	if (!sx_initialized(&table->lock))
		sx_init(&table->lock,table->name);
	list_prepend(&ipt_tables, table);

 unlock:
	up(&ipt_mutex);
	return ret;

 free_unlock:
	free(newinfo,M_IPTBUF);
	goto unlock;
}

void ipt_unregister_table(struct ipt_table *table)
{
	down(&ipt_mutex);
	LIST_DELETE(&ipt_tables, table);
	up(&ipt_mutex);

	/* Decrease module usage counts and free resources */
	IPT_ENTRY_ITERATE(table->private->entries, table->private->size,
			  cleanup_entry, NULL);
	free(table->private,M_IPTBUF);
}

struct ipt_error_target
{
	struct ipt_entry_target target;
	char errorname[IPT_FUNCTION_MAXNAMELEN];
};

struct ipt_error
{
	struct ipt_entry entry;
	struct ipt_error_target target;
};

struct __initial_table 
{
	struct ipt_replace repl;
	struct ipt_standard entries[3];
	struct ipt_error term;
};

/*
  *	So obscure definition:(
  *   the value is as following:
  *		struct ipt_replace repl {
  *			name:"filter"
  *			valid_hooks:FILTER_VALID_HOOKS
  *			num_entries:4
  *			size:sizeof(struct ipt_standard) * 3 + sizeof(struct ipt_error)
  *			hook_entry[NF_IP_LOCAL_IN]:0
  *			hook_entry[NF_IP_FORWARD]:sizeof(struct ipt_standard)
  *			hook_entry[NF_IP_LOCAL_OUT]:sizeof(struct ipt_standard) * 2
  *			underflow[NF_IP_LOCAL_IN]:0
  *			underflow[NF_IP_FORWARD]:sizeof(struct ipt_standard)
  *			underflow[NF_IP_LOCAL_OUT]:sizeof(struct ipt_standard) * 2
  *			num_counters:0
  *			counters:NULL
  *			entries:{}
  *		}
  *		struct ipt_standard entries[0] {
  *			struct ipt_entry entry {
  *				struct ipt_ip ip:0
  *				unsgigned int nfcache:0
  *				u_int16_t target_offset:sizeof(struct ipt_entry)
  *				u_int16_t next_offset:sizeof(struct ipt_standard)
  *				unsigned int comefrom:0
  *				struct ipt_counters counters:0
  *				unsigned char elems[0]:{}
  *			}
  *			struct ipt_standard_target target {
  *				struct ipt_entry_target target {
  *					u.user.target_size:IPT_ALIGN(sizeof(struct ipt_standard_target))
  *					u.user.name:""
  *					unsigned char data[0]:{}
  *				}
  *				int verdict:-NF_ACCEPT - 1
  *			}
  *		}
  */
static struct __initial_table initial_table
= { 
	/* struct ipt_replace repl */
	{ "filter", FILTER_VALID_HOOKS, 4,
      sizeof(struct ipt_standard) * 3 + sizeof(struct ipt_error),
      { [NF_IP_LOCAL_IN] 0,
	[NF_IP_FORWARD] sizeof(struct ipt_standard),
	[NF_IP_LOCAL_OUT] sizeof(struct ipt_standard) * 2 },
      { [NF_IP_LOCAL_IN] 0,
	[NF_IP_FORWARD] sizeof(struct ipt_standard),
	[NF_IP_LOCAL_OUT] sizeof(struct ipt_standard) * 2 },
      0, NULL, { } },
      /* struct ipt_standard entries[3] */
    {
	    /* LOCAL_IN */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } },
	    /* FORWARD */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } },
	    /* LOCAL_OUT */
	    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
		0,
		sizeof(struct ipt_entry),
		sizeof(struct ipt_standard),
		0, { 0, 0 }, { } },
	      { { { { IPT_ALIGN(sizeof(struct ipt_standard_target)), "" } }, { } },
		-NF_ACCEPT - 1 } }
    },
    /* ERROR */
    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0, 0 },
	0,
	sizeof(struct ipt_entry),
	sizeof(struct ipt_error),
	0, { 0, 0 }, { } },
      { { { { IPT_ALIGN(sizeof(struct ipt_error_target)), IPT_ERROR_TARGET } },
	  { } },
	"ERROR"
      }
    }
};

static struct ipt_table packet_filter;

/* The work comes in here from netfilter.c. */
static int
ipt_hook(void *arg,
	 struct mbuf **pskb,
	 struct ifnet *in,
	 int flags,
	 struct inpcb *inp)
{
	switch (ipt_do_table(pskb, NF_IP_LOCAL_IN, in, NULL, &packet_filter, arg)) {
		case NF_DROP:
			m_freem(*pskb);
		case NF_STOLEN:
			*pskb=NULL;
			break;
			
	}
	return 0;
}

static int
ipt_local_out_hook(void *arg,
		   struct mbuf **pskb,
		   struct ifnet *out,
		   int flags,
		   struct inpcb *inp)
{
	int ret=NF_ACCEPT;
	
	/* root is playing with raw sockets. */
	if ((*pskb)->m_len < sizeof(struct ip)
	    ||(mtod(*pskb, struct ip *))->ip_hl * 4 < sizeof(struct ip)) {
		if (net_ratelimit())
			log(LOG_DEBUG, "ipt_hook: happy cracking.\n");
	} else
		ret=ipt_do_table(pskb, NF_IP_LOCAL_OUT, NULL, out, &packet_filter, arg);
	switch (ret) {
		case NF_DROP:
			m_freem(*pskb);
		case NF_STOLEN:
			*pskb=NULL;
			break;
	}
	return 0;
}

static struct nf_hook_ops ipt_ops[]
= { { ipt_hook, PF_INET, PFIL_LOCAL_IN },
    { ipt_hook, PF_INET, PFIL_FORWARD },
    { ipt_local_out_hook, PF_INET, PFIL_LOCAL_OUT }
};

int nf_register_hook(struct nf_hook_ops *reg)
{
	struct pfil_head *pfh_inet;
	int flags=PFIL_WAITOK;

	if (reg->pf!=AF_INET&&reg->pf!=PF_INET) 
		return EPROTONOSUPPORT;
	flags|=reg->hookpos;
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if (pfh_inet == NULL)
		return ENOENT;
	return pfil_add_hook(reg->hook,NULL,flags,pfh_inet);
}

void nf_unregister_hook(struct nf_hook_ops *reg)
{
	struct pfil_head *pfh_inet;
	int flags=PFIL_WAITOK;
	
	if (reg->pf!=AF_INET&&reg->pf!=PF_INET) {
		log(LOG_ERR, "Error:unregister unkown nf_hook\n");
		return;
	}
	flags|=reg->hookpos;
	pfh_inet = pfil_head_get(PFIL_TYPE_AF, AF_INET);
	if (pfh_inet == NULL) {
		log(LOG_ERR, "Error:packet hook list not found\n");
		return;
	}
	if (pfil_remove_hook(reg->hook,NULL,flags,pfh_inet))
		log(LOG_ERR, "Error:remove hook error\n");
}

/* Functions to register sockopt ranges (exclusive). */
int 
nf_register_sockopt(struct nf_sockopt_ops *reg)
{
	struct list_head *i;
	int ret = 0;

	cv_init(&reg->cv,NULL);
	
	if (down_interruptible(&nf_sockopt_mutex) != 0)
		return EINTR;

	for (i = nf_sockopts.next; i != &nf_sockopts; i = i->next) {
		struct nf_sockopt_ops *ops = (struct nf_sockopt_ops *)i;
		if (ops->pf == reg->pf
		    && (overlap(ops->set_optmin, ops->set_optmax, 
				reg->set_optmin, reg->set_optmax)
			|| overlap(ops->get_optmin, ops->get_optmax, 
				   reg->get_optmin, reg->get_optmax))) {
			NFDEBUG("nf_sock overlap: %u-%u/%u-%u v %u-%u/%u-%u\n",
				ops->set_optmin, ops->set_optmax, 
				ops->get_optmin, ops->get_optmax, 
				reg->set_optmin, reg->set_optmax,
				reg->get_optmin, reg->get_optmax);
			ret = EBUSY;
			goto out;
		}
	}

	list_add(&reg->list, &nf_sockopts);
out:
	up(&nf_sockopt_mutex);
	return ret;
}

void nf_unregister_sockopt(struct nf_sockopt_ops *reg)
{
	/* No point being interruptible: we're probably in cleanup_module() */

	down(&nf_sockopt_mutex);
 restart:
	if (reg->use != 0) {
		/* To be woken by nf_sockopt call... */
		/* FIXME: Stuart Young's name appears gratuitously. */
		reg->cleanup_task = curthread;
		cv_wait(&reg->cv,&nf_sockopt_mutex);
		goto restart;
	}
	list_del(&reg->list);
	up(&nf_sockopt_mutex);
}

/* Call get/setsockopt() */
static int nf_sockopt(struct socket *sk, int pf, int val, 
		      char *opt, int *len, int get)
{
	struct list_head *i;
	struct nf_sockopt_ops *ops;
	int ret;

	if (down_interruptible(&nf_sockopt_mutex) != 0)
		return EINTR;

	for (i = nf_sockopts.next; i != &nf_sockopts; i = i->next) {
		ops = (struct nf_sockopt_ops *)i;
		if (ops->pf == pf) {
			if (get) {
				if (val >= ops->get_optmin
				    && val < ops->get_optmax) {
					ops->use++;
					up(&nf_sockopt_mutex);
					ret = ops->get(sk, val, opt, len);
					goto out;
				}
			} else {
				if (val >= ops->set_optmin
				    && val < ops->set_optmax) {
					ops->use++;
					up(&nf_sockopt_mutex);
					ret = ops->set(sk, val, opt, *len);
					goto out;
				}
			}
		}
	}
	up(&nf_sockopt_mutex);
	return ENOPROTOOPT;
	
 out:
	down(&nf_sockopt_mutex);
	ops->use--;
	
	if (ops->cleanup_task)
		cv_signal(&ops->cv);
	up(&nf_sockopt_mutex);
	return ret;
}

int nf_setsockopt(struct socket *sk, int pf, int val, char *opt,
		  int len)
{
	return nf_sockopt(sk, pf, val, opt, &len, 0);
}

int nf_getsockopt(struct socket *sk, int pf, int val, char *opt, int *len)
{
	return nf_sockopt(sk, pf, val, opt, len, 1);
}

static int forward = NF_ACCEPT;

/*
  *	Linux iptables filter table initialize
  */
static int linux_iptables_filter_init(void)
{
	int ret;

	if (forward < 0 || forward > NF_MAX_VERDICT) {
		log(LOG_ERR, "iptables forward must be 0 or 1\n");
		return EINVAL;
	}

	/* Entry 1 is the FORWARD hook */
	initial_table.entries[1].target.verdict = -forward - 1;

      bzero(&packet_filter,sizeof(packet_filter));
	strcpy(packet_filter.name,"filter");
	packet_filter.table=&initial_table.repl;
	packet_filter.valid_hooks=FILTER_VALID_HOOKS;
	sx_init(&packet_filter.lock,"packet filter lock");

	/* Register table */
	ret = ipt_register_table(&packet_filter);
	if (ret)
		return ret;

	/* Register hooks */
	ret = nf_register_hook(&ipt_ops[0]);
	if (ret) {
		goto cleanup_table;
	}
	
	ret = nf_register_hook(&ipt_ops[1]);
	if (ret) {
		goto cleanup_hook0;
	} 
	
	ret = nf_register_hook(&ipt_ops[2]);
	if (ret < 0)
		goto cleanup_hook1;

	return ret;

 cleanup_hook1:
	nf_unregister_hook(&ipt_ops[1]);
 cleanup_hook0:
	nf_unregister_hook(&ipt_ops[0]);
 cleanup_table:
	ipt_unregister_table(&packet_filter);

	return ret;
}

/*
  *	Linux iptables initialize
  */
static int linux_iptables_init(void)
{
	int ret;

	INIT_LIST_HEAD(&ipt_tables);
	INIT_LIST_HEAD(&ipt_target);
	INIT_LIST_HEAD(&ipt_match);
	
	mtx_init(&ipt_mutex,"iptables lock",NULL,MTX_DEF);
	
	/* Noone else will be downing sem now, so we won't sleep */
	down(&ipt_mutex);
	list_append(&ipt_target, &ipt_standard_target);
	list_append(&ipt_target, &ipt_error_target);
	list_append(&ipt_match, &tcp_matchstruct);
	list_append(&ipt_match, &udp_matchstruct);
	list_append(&ipt_match, &icmp_matchstruct);
	up(&ipt_mutex);

	/* Register setsockopt */
	ret = nf_register_sockopt(&ipt_sockopts);
	if (ret < 0) {
		printf("Unable to register sockopts.\n");
		return ret;
	}
	
	ret = linux_iptables_filter_init();
	if (ret) {
		printf("Linux iptables filter init error\n");
		return ret;
	}
	
	printf("ip_tables: (C) 2000-2002 Netfilter core team\n");
	return 0;
}
/*
  *	Netfilter initialize,called by ip_init()(ip_input.c)
  */
int linux_netfilter_init(void)
{
	INIT_LIST_HEAD(&nf_sockopts);
	mtx_init(&nf_sockopt_mutex,"linux netfilter lock",NULL,MTX_DEF);
	linux_iptables_init();
	
	return 0;
}

void linux_netfilter_fini(void)
{
	nf_unregister_sockopt(&ipt_sockopts);
	mtx_destroy(&ipt_mutex);
	mtx_destroy(&nf_sockopt_mutex);
}
